"use strict";

const EventEmitter = require('events');
const Promise      = require('bluebird');
const debug        = require('debug')('eventbus:store');
const pg           = require('pg');
const tables       = require('./tables');

const INIT_RETRY_INTERVAL = 5; // in seconds

class Store extends EventEmitter {

    constructor(microServiceName, options) {
        super();
        this.microServiceName = microServiceName;
        this.host             = options.host;
        this.database         = options.database;
        this.user             = options.user;
        this.password         = options.password;
        this.readyPromise     = new Promise((resolve, reject) => {
            this.readyDeferred = {
                resolve: () => {
                    resolve();
                },
                reject: () => {
                    reject();
                }
            };
        });
    }

    quit() {
        // TODO implement
    }

    init() {
        const config = {
            user: this.user,
            database: this.database,
            password: this.password,
            host: this.host,
            port: 5432,
            max: 10,
            idleTimeoutMillis: 30000, // how long a client is allowed to remain idle before being closed
        };
        this.pool    = new pg.Pool(config);

        this.pool.on('error', (err) => {
            this.emit('error', err);
        });

        this._assertSystemTables()
            .then(() => {
                debug(`ready`);
                this.emit('ready');
            })
            .catch((err) => {
                this.emit('error', err);
                debug(`initialization failed trying again in ${INIT_RETRY_INTERVAL} seconds`);
                setTimeout(() => {
                    this.init();
                }, INIT_RETRY_INTERVAL * 1000);
            })
            .done();
    }

    _assertSystemTables() {
        return Promise.promisify(this.pool.query.bind(this.pool))(tables)
            .then(() => {
                debug('system tables asserted');
            });
    }

    save(eventName, event) {
        let data = [
            { field: 'event', value: eventName },
            { field: 'props', value: event.props },
            { field: 'aggregate_id', value: event.aggregateId },
            { field: 'timestamp', value: new Date() },
            { field: 'publisher', value: this.microServiceName }
        ];

        if (event.accountId) {
            data.push({ field: 'account_id', value: event.accountId });
        }

        if (event.userId) {
            data.push({ field: '"user_id"', value: event.userId });
        }

        let fieldString  = '';
        let valuesString = '';
        let dataArr      = [];
        data.forEach((d, index) => {
            if (!index) {
                fieldString += `${d.field}`;
                valuesString += `$${index + 1}`;
            } else {
                fieldString += `, ${d.field}`;
                valuesString += `, $${index + 1}`;
            }
            dataArr.push(d.value);
        });

        const storeQuery = `INSERT INTO eventstore.events (${fieldString}) VALUES (${valuesString})`;
        return Promise.promisify(this.pool.query.bind(this.pool))(storeQuery, dataArr);
    }

    getAllAggregateEvents(aggregateId) {
        const getQuery = `SELECT * FROM eventstore.events WHERE aggregate_id = '${aggregateId}'`;
        return Promise.promisify(this.pool.query.bind(this.pool))(getQuery);
    }
}


module.exports = Store;
