'use strict';

const EventEmitter = require('events');
const Promise      = require('bluebird');
const debug        = require('debug')('eventbus:main');
const uuid         = require('uuid');
const Bus          = require('./Bus');
const Store        = require('./Store');

//TODO
// 1. queue published messages in case of failure in the connection
// 2. re subscribe in case of reconnection
// 3. allow to control how and when to ack a subscribed event

class EventBus extends EventEmitter {

    constructor() {
        super();
        this.busReady     = false;
        this.storeReady   = false;
        this.readyPromise = new Promise((resolve, reject) => {
            this.readyDeferred = {
                resolve: () => {
                    resolve();
                },
                reject: () => {
                    reject();
                }
            };
        });
        this.on('error', () => {});
    }

    quit() {
        this.bus.quit();
        this.store.quit();
        // TODO emit a quit event once cleanly everything was closed
    }

    init(options = {}) {
        this.microServiceName = options.microServiceName || `eventBus_client_${uuid()}`;

        this.bus   = new Bus(this.microServiceName, options.bus);
        this.store = new Store(this.microServiceName, options.store);

        // wait for both the bus and the store to send their ready event
        this.bus.on('ready', () => {
            debug('bus is ready');
            this.busReady = true;
            this._emitReady();
        });

        this.store.on('ready', () => {
            debug(`store is ready. registered as: ${this.microServiceName}`);
            this.storeReady = true;
            this._emitReady();
        });

        this.bus.on('error', (err) => {
            debug(`bus error: ${err.toString()}`);
            this.emit('error', err);
        });

        this.store.on('error', (err) => {
            debug(`store error: ${err.toString()}`);
            this.emit('error', err);
        });

        this.bus.init();
        this.store.init();
    }

    _emitReady() {
        if (this.busReady && this.storeReady) {
            this.readyDeferred.resolve();
            debug('ready');
            this.emit('ready');
        }
    }

    publish(eventName, msg) {
        return this.readyPromise
            .then(() => {
                return this.store.save(eventName, msg);
            })
            .then(() => {
                return this.bus.publish(eventName, msg);
            });
    }

    subscribe(eventName, handler) {
        return this.readyPromise
            .then(() => {
                return this.bus.subscribe(eventName, handler);
            });
    }

    getAllAggregateEvents(aggregateId) {
        return this.readyPromise
            .then(() => {
                return this.store.getAllAggregateEvents(aggregateId);
            });
    }
}

module.exports = new EventBus();
